/*
** main.c
*/

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>

#include <main.h>

typedef enum	optval_e
{
  HELP = 'h',
  VERSION = 'v',
  LIST = 'l',
  FIND = 'f',
  TEST = 't',
  SET = 's',
  INIT = 'i',
  CORRCKS = 'c',
  BIOS = 'b',
}		optval_t;
static char		*optshort = "hvlb:f:t:s:ic";
static struct option	optlong[] =
{
  {"help",		no_argument,		0,	HELP},
  {"version",		no_argument,		0,	VERSION},
  {"list",		no_argument,		0,	LIST},
  {"find",		required_argument,	0,	FIND},
  {"test",		required_argument,	0,	TEST},
  {"set",		required_argument,	0,	SET},
  {"init",		no_argument,		0,	INIT},
  {"corrcks",		no_argument,		0,	CORRCKS},
  {"bios",		required_argument,	0,	BIOS},
  {0,			0,			0,	0},
};

/*
** All supported bioses must be included in this array
*/
bios_t		bioses[] =
{
  {
    "awdmod",
    "Award Modular BIOS v6.00PG, and probably many other versions\n"
    "Passwords are case-sensitive\n"
    "Successfuly tested on these motherboards: MSI-6167, MSI-6178\n",
    1, 8, ' ', '~',
    awdmod_pwd_check,
    awdmod_pwd_set_onoff
  },
  {
    "awdmedal",
    "Award Medallion BIOS v6.0, and probably many other versions\n"
    "Passwords are not case-sensitive\n",
    1, 8, 'A', 'Z',
    awdmedal_pwd_check,
    NULL
  },
};

/*
** Usage
*/
void		usage(const char *s)
{
  fprintf
    (stderr,
     "Usage:\n"
     "\t%s {INFO | ACTION} [BIOS] \n"
     "\nINFO options available:\n"
     "\t-h --help            display this help and exit\n"
     "\t-v --version         display program version and exit\n"
     "\t-l --list            list all supported bioses and exit\n"
     "\nACTION options available:\n"
     "\t-f --find=<method>   find valid passwords (<method>: brute, stdin)\n"
     "\t-t --test=<password> test the validity of password <password>\n"
     "\t-s --set=<state>     enable/disable the password (<state>: on, off)\n"
     "\t-i --init            initialize nvram content\n"
     "\t-c --corrcks         correct nvram checksum\n"
     "\nThe BIOS option (required by --find, --test and --set) is:\n"
     "\t-b --bios=<biostype> specify the bios type\n"
     , s);
}

/*
** Version
*/
void		version(const char *s)
{
    fprintf(stderr, "%s %s\n", s, PROGRAM_VERSION);
}

/*
** List all supported bioses
*/
void		list(void)
{
  unsigned	i;

  printf("List of all supported bioses and their description:\n\n");
  for (i = 0; bioses[i].name; i++)
      printf("* biostype '%s' (--set option %ssupported)\n%s\n",
	     bioses[i].name,
	     bioses[i].pwd_set_onoff ? "" : "not ",
	     bioses[i].desc);
}

/*
** Entry point
*/
int		main(int argc, char **argv)
{
  int		bios = -1;
  unsigned	do_find = 0;
  method_t	method = M_BRUTE;
  unsigned	do_test = 0;
  char		*pwd = NULL;
  unsigned	do_set = 0;
  int		state = 0;
  unsigned	do_init = 0;
  unsigned	do_corrcks = 0;
  int		fd;
  int		c;
  unsigned	i;

  /*
  ** Args processing
  */
  while (-1 != (c = getopt_long(argc, argv, optshort, optlong, 0)))
    {
      switch (c)
	{
	case HELP:
	  usage(argv[0]), exit(EXIT_SUCCESS);
	case VERSION:
	  version(argv[0]), exit(EXIT_SUCCESS);
	case LIST:
	  list(), exit(EXIT_SUCCESS);
	case BIOS:
	  for (i = 0; bioses[i].name; i++)
	    if (!strcmp(optarg, bioses[i].name))
	      bios = i;
	  if (bios == -1)
	    fprintf(stderr, "%s: invalid bios\n", optarg),
	      exit(EXIT_FAILURE);
	  break;
	case FIND:
	  do_find = 1;
	  if (!strcmp(optarg, "stdin"))
	    method = M_STDIN;
	  else if (!strcmp(optarg, "brute"))
	    method = M_BRUTE;
	  else
	    fprintf(stderr, "%s: invalid method\n", optarg),
	      exit(EXIT_FAILURE);
	  break;
	case TEST:
	  do_test = 1;
	  pwd = optarg;
	  break;
	case SET:
	  do_set = 1;
	  if (!strcmp(optarg, "on"))
	    state = 1;
	  else if (!strcmp(optarg, "off"))
	    state = 0;
	  else
	    fprintf(stderr, "%s: invalid state\n", optarg),
	      exit(EXIT_FAILURE);
	  break;
	case INIT:
	  do_init = 1;
	  break;
	case CORRCKS:
	  do_corrcks = 1;
	  break;
	default:
	  usage(argv[0]), exit(EXIT_FAILURE);
	}
    }
  /*
  ** Args checking
  */
  if (1 != do_find + do_test + do_set + do_init + do_corrcks)
    {
      fprintf(stderr, "There must be 1 INFO option or 1 ACTION option\n"
	      "Try '%s --help'\n", argv[0]);
      exit(EXIT_FAILURE);
    }
  if ((do_find || do_test || do_set) && (bios == -1))
    {
      fprintf(stderr, "The ACTION option you specified (--find, --test or "
	      "--set) require the --bios option\n"
	      "Try '%s --help'\n", argv[0]);
      exit(EXIT_FAILURE);
    }
  /*
  ** Open the device
  */
  if (-1 == (fd = open(NVRAM_DEVICE, O_RDWR, 0)))
    {
      perror(NVRAM_DEVICE);
      if (errno == ENODEV)
	  fprintf(stderr, "Your Linux kernel does not support the /dev/nvram "
		  "device; it has to be configured and compiled with "
		  "CONFIG_NVRAM\n");
      exit(EXIT_FAILURE);
    }
  /*
  ** Before doing anything, check the validity of the nvram checksum
  ** (unless we are going to correct it with --corrcks).
  ** drivers/char/nvram.c says:
  **   "Checksums over the NVRAM contents are managed by this driver.
  **    In case of a bad checksum, reads and writes return -EIO."
  */
  if (!do_corrcks && 1 != read(fd, &c, 1))
    {
      perror("read");
      if (errno == EIO)
	  fprintf(stderr, "The nvram checksum is bad; correct it with "
		  "the --corrcks option, then try again\n");
      exit(EXIT_FAILURE);
    }
  /*
  ** Now, the real work
  */
  if (do_find)
      pwd_find(fd, bioses + bios, method);
  if (do_test)
      pwd_test(fd, bioses + bios, pwd);
  if (do_set)
      pwd_set(fd, bioses + bios, state);
  if (do_init)
      nvram_init(fd);
  if (do_corrcks)
      nvram_corrcks(fd);
  return 0;
}
